"""Search sub command to emulate Netbox main search bar."""


from pynetbox.core.query import Request
from pynetbox.core.endpoint import response_loader
from nbcli.commands.base import BaseSubCommand
from nbcli.core.utils import app_model_by_loc
from nbcli.views.tools import nbprint


class SearchSubCommand(BaseSubCommand):
    """Search Netbox objects with the given searchterm.

    The List of search objects can be modified in:
    $CONF_DIR/user_config.yml
    """

    name = 'search'
    parser_kwargs = dict(help='Search Netbox Objects')

    def setup(self):
        """Add parser arguments to search sub command."""
        self.parser.add_argument('obj_type',
                                 type=str,
                                 nargs='?',
                                 help='Object type to search')

        self.parser.add_argument('searchterm',
                                 help='Search term')

    def run(self):
        """Run a search of Netbox objects and show a table view of results.

        Usage Examples:

        - Search all object types for 'server1':
          $ nbcli search server1

        - Search the interface object type for 'eth 1':
          $ nbcli search interface 'eth 1'
        """
        if hasattr(self.netbox.nbcli.conf, 'nbcli') and \
                ('search_objects' in self.netbox.nbcli.conf.nbcli.keys()):
            self.search_objects = self.netbox.nbcli.conf.nbcli['search_objects']
        else:
            self.search_objects = ['provider',
                                  'circuit',
                                  'site',
                                  'rack',
                                  'location',
                                  'device_type',
                                  'device',
                                  'virtual_chassis',
                                  'cable',
                                  'power_feed',
                                  'vrf',
                                  'aggregate',
                                  'prefix',
                                  'address',
                                  'vlan',
                                  'secret',
                                  'tenant',
                                  'cluster',
                                  'virtual_machine']

        def search(ep):
            req = Request(filters=dict(q=self.args.searchterm),
                          base=ep.url,
                          token=ep.token,
                          session_key=ep.session_key,
                          http_session=ep.api.http_session)

            rep = req._make_call(add_params=dict(limit=15))

            result = list()

            if rep.get('results'):
                result = response_loader(rep['results'], ep.return_obj, ep)

            return result

        self.nbprint = nbprint

        if self.args.obj_type:
            modellist = [self.args.obj_type]
        else:
            modellist = self.search_objects

        result_count = 0

        print('')
        for obj_type in modellist:
            model = app_model_by_loc(self.netbox, obj_type)
            result = search(model)
            full_count = model.count(self.args.searchterm)
            if len(result) > 0:
                result_count += 1
                print('{}\n{}'.format(obj_type.title(), '=' * len(obj_type)))
                self.nbprint(result)
                if len(result) < full_count:
                    print('*** See all {} results: '.format(full_count) +
                          "'$ nbcli filter {} {}' ***".
                          format(obj_type, self.args.searchterm))
                print('')
        if result_count == 0:
            self.logger.warning('No results found')
            print('')



from nbcli.views.tools import BaseView

class DcimLocationsView(BaseView):

    def table_view(self):

        self.add_col('Name', self.get_attr('name'))
        self.add_col('Site', self.get_attr('site'))
        self.add_col('Racks', self.get_attr('rack_count'))
        self.add_col('Description', self.get_attr('description'))



import yaml
from nbcli.core.utils import app_model_by_loc
from nbcli.commands.base import BaseSubCommand
from nbcli.commands.tools import NbArgs


class Upsert():

    def __init__(self, netbox, logger, model, data,
                 res=None,
                 dr=False,      # dry run
                 ud=False,      # update only
                 parent=None):

        assert (parent == None) or isinstance(parent, Upsert)

        if isinstance(data, list):
            for d in data:
                Upsert(netbox, logger, model, d, dr=dr, ud=ud, parent=parent)
            return

        self.netbox = netbox
        self.logger = logger
        self.model = model
        self.data = data
        self.dr = dr
        self.ud = ud
        self.parent = parent

        self.res = res or netbox.nbcli.rm.get(self.model.split(':')[0])
        self.ep = app_model_by_loc(self.netbox, self.res.model)
        self.args = None
        self.obj = None
        self.children = list()  # list of tuples (model, data, res)

        assert self.res

        self.obj = None

        self.proc_model()

        self.action()

        self.proc_children()

    def _add_parent_arg(self):
        if self.parent:
            res = self.parent.res.get(self.res.model) or self.parent.res
            self.args.apply_res([self.parent.obj], res)


    def proc_model(self):

        gp = True
        kws = None

        if self.model.endswith('^'):
            gp = False
            self.model = self.model[:-1]
    
        if ':' in self.model:
            self.args = NbArgs(self.netbox)
            alias, kws = self.model.split(':', 1)
            if gp:
                self._add_parent_arg()

            nba, self.obj = self.args.resolve(alias, kws, kwargs=self.args.kwargs, res=self.res)
            if self.obj:
                assert len(self.obj) == 1
                self.obj = self.obj[0]
                self.args = NbArgs(self.netbox, action='patch')
                if not gp:
                    self._add_parent_arg()
            else:
                self.obj = None
                self.args = NbArgs(self.netbox, action='post')
                self.args.proc(*nba.kwargs.items())
        else:
            self.args = NbArgs(self.netbox, action='post')

        self._add_parent_arg()

    def action(self):

        for key, value in self.data.items():
            self.proc_data_items(key, value)
        
        if self.obj:
            self.logger.info('Updating %s with data: %s',
                             str(self.obj), str(self.args.kwargs))
            self.obj.update(self.args.kwargs)
        else:
            self.logger.info('Creating %s with data: %s',
                             self.res.alias, str(self.args.kwargs))
            self.obj = self.ep.create(**self.args.kwargs)

    def proc_data_items(self, key, value, create=False):

        rstr = key.split(':')[0]
        res = self.res.get(rstr) or self.netbox.nbcli.rm.get(rstr)

        if res:
            if  (':' in key) and (value is None):
                self.children.append((key, {}, res))
            elif isinstance(value, (dict, list)):
                self.children.append((key, value, res))
            else:
                self.args.resolve(key, value, res=res)
        else:
            self.args.update(key, value)

    def proc_children(self):

        for child in self.children:

            model, data, res = child

            Upsert(self.netbox, self.logger, model, data,
                   res=res,
                   dr=self.dr,
                   ud=self.ud,
                   parent=self)


class CreateSubCommand(BaseSubCommand):
    """Create and/or Update objects defined in YAML file."""

    name = 'create'
    parser_kwargs = dict(help='Create/Update objects with YAML file.')
    default_loglevel = 20

    def setup(self):

        self.parser.add_argument('file',
                                 type=str,
                                 help='YAML file.')
        self.parser.add_argument('--dr', '--dry-run',
                                 action='store_true',
                                 help='Dry run.')
        self.parser.add_argument('-u', '--update-only',
                                 action='store_true',
                                 help='Do not create object not found '+ \
                                      'with the lookup key schema.')

    def run(self):
        """Run command.

        See documentation for YAML file reference and examples.
        https://nbcli.readthedocs.io/en/release/create.html

        Usage Examples:

        - Create/Update objects defined in YAML file
          $ nbcli create file.yml
        """

        with open(self.args.file) as fh:
            data_stream = yaml.safe_load_all(fh.read())

        for data in data_stream:
            for key, value in data.items():
                assert self.netbox.nbcli.rm.get(key.split(':')[0])
                Upsert(self.netbox, self.logger, key, value,
                       dr=self.args.dr,
                       ud=self.args.update_only,
                       parent=None)



"""Base sub-command and tools for commands."""


import argparse
from argparse import RawTextHelpFormatter
import functools
import logging
import sys
from pydoc import getdoc
from textwrap import dedent
from nbcli.core.config import get_session
from nbcli.views.tools import nbprint


def get_common_parser():
    """Create parser for handling verbose options."""
    common_parser = argparse.ArgumentParser(add_help=False)
    common_parser.add_argument('-v', '--verbose',
                               action='count',
                               help='Show more logging messages')
    common_parser.add_argument('-q', '--quiet',
                               action='count',
                               help='Show fewer logging messages')
    return common_parser


def get_view_parser():
    """Create parser for handling view options."""
    view_parser = argparse.ArgumentParser(add_help=False)
    view_parser.add_argument('--view', type=str, help='Output view.',
                             choices=['table', 'detail', 'json'],
                             default='table')
    view_parser.add_argument('--view-model',
                             type=str,
                             help='View model to use')
    view_parser.add_argument('--cols', nargs='*',
                             help="Custom columns for table output.")
    view_parser.add_argument('--nh', '--no-header',
                             action='store_true',
                             help='Disable header row in results')
    return view_parser


class BaseSubCommand():
    """Base sub-command to build commands from."""

    name = 'base'
    parser_kwargs = dict(help='')
    view_options = False
    default_loglevel = logging.WARNING

    def __init__(self, subparser):
        """Add sub-command parser to subparser object."""
        if 'parents' in self.parser_kwargs.keys():
            assert isinstance(self.parser_kwargs['parents'], list)
            self.parser_kwargs['parents'].append(get_common_parser())
        else:
            self.parser_kwargs['parents'] = [get_common_parser()]

        if self.view_options:
            self.parser_kwargs['parents'].append(get_view_parser())

        self.name = self.name.lower()

        if 'formatter_class' not in self.parser_kwargs.keys():
            self.parser_kwargs['formatter_class'] = RawTextHelpFormatter

        if 'description' not in self.parser_kwargs.keys():
            self.parser_kwargs['description'] = dedent(getdoc(self))

        if 'epilog' not in self.parser_kwargs.keys():
            self.parser_kwargs['epilog'] = dedent(getdoc(self.run))

        # try to resolve conflicting sub_command names
        if self.name in dict(subparser._get_kwargs())['choices'].keys():
            if self.__module__.split('.')[0] == 'user_commands':
                self.name = 'user_{}'.format(self.name)
            else:
                prefix = self.__module__.split('.')[0].replace('nbcli_', '')
                self.name = '{}_{}'.format(prefix, self.name)

        self.parser = subparser.add_parser(self.name, **self.parser_kwargs)
        self.parser.set_defaults(func=self._pre_run_)
        self.setup()

    def _pre_run_(self, args):

        self.args = args
        self.logger = logging.getLogger('nbcli.'+self.name)
        self._set_log_level_()
        try:
            self.netbox = get_session()
            if self.view_options:
                nbopts = dict(view=self.args.view,
                              view_model=self.args.view_model,
                              cols=self.args.cols,
                              disable_header=self.args.nh)

                self.nbprint = functools.partial(nbprint, **nbopts)
            self.logger.debug(args)

            self.run()
        except Exception as e:
            from traceback import print_tb
            self.logger.critical('%s: %s', type(e).__name__, str(e))

            if self.logger.parent.level <= 10:
                print_tb(e.__traceback__)
            if self.logger.parent.level == 1:
                try:
                    from ipdb import post_mortem
                except:
                    from pdb import post_mortem
                print("\n*** Entering debuger! " + \
                      "(type '?' for help, or 'q' to quit) ***\n")
                post_mortem()

            sys.exit(1)
        except KeyboardInterrupt:
            self.logger.warning('Interrupted by user.')
            sys.exit(0)

    def _set_log_level_(self):
        """Set the loglevel based on the arguments passed."""
        level = self.default_loglevel

        if self.args.quiet:
            level = level + (10 * self.args.quiet)
            if level > 100:
                level = 100

        if self.args.verbose:
            level = level - (10 * self.args.verbose)
            if level < 1:
                level = 1

        self.logger.parent.setLevel(level)

    def setup(self):
        """Define additional parser options for sub command."""
        pass

    def run(self):
        """Execute subcommand."""
        pass



from nbcli.views.tools import BaseView

class ExtrasObjectChangesView(BaseView):

    def table_view(self):

        self.add_col('Time', self.get_attr('time').split('.')[0].replace('T', ' '))
        self.add_col('User', self.get_attr('user.username'))
        self.add_col('Action', self.get_attr('action'))
        self.add_col('Type', self.get_attr('changed_object_type').split('.')[-1])
        self.add_col('Object', self.get_attr('changed_object'))
        self.add_col('RequestID', self.get_attr('request_id'))



from nbcli.views.tools import BaseView

class IpamIpAddressesView(BaseView):

    def table_view(self):

        self.add_col('IP Address', self.get_attr('address'))
        self.add_col('Vrf', self.get_attr('vrf'))
        self.add_col('Status', self.get_attr('status'))
        self.add_col('Role', self.get_attr('role'))
        self.add_col('Tenant', self.get_attr('tenant'))
        self.add_col('Parent', self.get_attr('interface.device'))
        self.add_col('Interface', self.get_attr('interface'))
        self.add_col('DNS Name', self.get_attr('dns_name'))
        self.add_col('Description', self.get_attr('description'))



"""Sub Command to set up bare nbcli user directory."""

from nbcli.commands.base import BaseSubCommand
from nbcli.core.utils import app_model_by_loc, rend_table
from nbcli.views.tools import view_name


class ModelsSubCommand(BaseSubCommand):
    """View Information on nbcli Models."""

    name = 'models'
    parser_kwargs = dict(help='Display information on Supported Models.')

    def setup(self):
        """Additional arguments for model command."""
        self.parser.add_argument('model',
                                 nargs='?',
                                 help='Display detailed info on given model.')

    def run(self):
        """View Information on nbcli Models.

        Example Usage:

        - List all supported models
          $ nbcli models

        - Show information of device model
          $ nbcli models device
        """
        if not self.args.model:

            header = ['Model', 'Lookup', 'Endpoint']

            table = [header]

            for res in self.netbox.nbcli.rm:
                loc = res.model.replace('.', '/').replace('_', '-')
                row = [res.alias, res.lookup, loc]
                table.append(row)

            print(rend_table(table))

        else:

            res = self.netbox.nbcli.rm.get(self.args.model)
            if res:
                ep = app_model_by_loc(self.netbox, res.model)

                disp = f'\nModel: {res.alias}\n' + \
                       f'Lookup: {res.lookup}\n' + \
                       f'View Name: {view_name(ep)}\n' + \
                       f'API Endpoint: {ep.url}\n'

                print(disp)

            else:
                self.logger.warning("Unsupported model: '%s'",
                                    self.args.model)



"""Load views used by nbprint."""

from nbcli.views.tools import BaseView                          # noqa: F401
from nbcli.views.device import DcimDevicesView                  # noqa: F401
from nbcli.views.interfaces import DcimInterfacesView           # noqa: F401
from nbcli.views.racks import DcimRacksView, DcimRUsView        # noqa: F401
from nbcli.views.locations import DcimLocationsView             # noqa: F401
from nbcli.views.sites import DcimSitesView                     # noqa: F401
from nbcli.views.aggregates import IpamAggregatesView           # noqa: F401
from nbcli.views.ip_addresses import IpamIpAddressesView        # noqa: F401
from nbcli.views.object_changes import ExtrasObjectChangesView  # noqa: F401



from nbcli.views.tools import BaseView

class DcimRacksView(BaseView):

    def table_view(self):

        self.add_col('Name', self.get_attr('name'))
        self.add_col('Site', self.get_attr('site'))
        self.add_col('Group', self.get_attr('group'))
        self.add_col('Status', self.get_attr('status'))
        self.add_col('Facility ID', self.get_attr('facility_id'))
        self.add_col('Tenant', self.get_attr('tenant'))
        self.add_col('Role', self.get_attr('role'))
        self.add_col('Height', self.get_attr('u_height'))
        self.add_col('Devices', self.get_attr('device_count'))


class DcimRUsView(BaseView):

    def table_view(self):

        self.add_col('Name', self.get_attr('name'))
        self.add_col('Device', self.get_attr('device'))
        self.add_col('Role', self.get_attr('device.device_role'))
        self.add_col('Type', self.get_attr('device.device_type'))
        self.add_col('Serial', self.get_attr('device.serial'))



from nbcli.views.tools import BaseView

class IpamAggregatesView(BaseView):

    def table_view(self):

        self.add_col('Aggregate', self.get_attr('aggregate'))
        self.add_col('RIR', self.get_attr('rir'))
        self.add_col('Added', self.get_attr('date_added'))
        self.add_col('Description', self.get_attr('description'))



"""Define function for loading user extensions and plugins."""

import importlib
import pkgutil
import sys
from nbcli import logger
from nbcli.core.utils import get_nbcli_dir


def load_extensions():
    """Load user extensions and plugins."""
    sys.path.append(str(get_nbcli_dir().joinpath('user_extensions')))

    extensions = list()

    for _, name, _ in pkgutil.iter_modules():
        if name.startswith('nbcli_'):
            extensions.append(name)

    extensions.append('user_views')
    extensions.append('user_commands')

    sys.dont_write_bytecode = True

    for ext in extensions:

        try:
            importlib.import_module(ext)
            logger.info('%s loaded.', ext)
        except Exception as e:
            logger.error('Error loading %s!', ext)
            logger.error('%s: %s', type(e).__name__, str(e))
            if 0 < logger.level <= 10:
                raise e

    sys.dont_write_bytecode = True



from nbcli.views.tools import BaseView

class DcimInterfacesView(BaseView):

    def table_view(self):

        self.add_col('Parent', self.get_attr('device'))
        self.add_col('Name', self.get_attr('name'))
        self.add_col('Enabled', self.get_attr('enabled'))
        self.add_col('Type', self.get_attr('type'))
        self.add_col('Description', self.get_attr('description'))
        self.add_col('cable', self.get_attr('cable.id'))



"""Define Classes and Functions for use by views."""

import json
from collections import OrderedDict
from pynetbox.core.response import Record
from nbcli.core.utils import app_model_loc, is_list_of_records, view_name, rend_table


class BaseView():
    """Base (default) view used to create other netbox object views."""

    def __init__(self, obj, cols=list()):
        """Populate _view based on defined table_view of passed cols."""
        assert isinstance(obj, Record)
        self._obj = obj
        self._view = OrderedDict()
        if cols and isinstance(cols, list) and (len(cols) > 0):
            for col in cols:
                if isinstance(col, str):
                    col = (col, col)
                assert isinstance(col, tuple) and (len(col) == 2)
                self.add_col(col[0], self.get_attr(col[1]))
        else:
            self.table_view()

    @property
    def obj(self):
        """Return object associated with view instance."""
        return self._obj

    @property
    def view(self):
        """Return populated OrderedDict."""
        return self._view

    def add_col(self, header, value):
        """Convert header/value to string and add to view."""
        if str(value).lower() in ['none', '']:
            value = '-'

        self._view[str(header)] = str(value)

    def get_attr(self, attribute):
        """Resolve attribute of assigned object.

        Return the value, or None if it does not exist.
        """
        assert isinstance(attribute, str)

        def get_val(obj, key):

            if isinstance(obj, dict) and (key in obj):
                return obj[key]
            elif isinstance(obj, list) and key.isdigit() and (len(obj) > int(key)):
                return obj[int(key)]
            else:
                return None

        obj = self.obj

        for attr in attribute.split('.'):

            keys = attr.split(':')[1:]
            attr = attr.split(':')[0]

            if hasattr(obj, attr.lower()):
                obj = getattr(obj, attr.lower())
                for key in keys:
                    obj = get_val(obj, key)
            else:
                return None

        return obj

    def table_view(self):
        """Define headers and values for table view of object."""
        self.add_col('ID', self.get_attr('id'))
        self.add_col(view_name(self.obj).replace('View', ''),
                     str(self.obj))

    def detail_view(self):
        """Define detail view of object."""
        lines = list()
        for attr in dict(self.obj).keys():
            lines.append(attr + ': ' + str(self.get_attr(attr)))
        return '\n'.join(lines)

    def items(self):
        """Return items of OrderedDict."""
        return self.view.items()

    def keys(self):
        """Return keys of OrderedDict."""
        return self.view.keys()

    def values(self):
        """Get values of OrderedDict."""
        return self.view.values()

    def __iter__(self):
        """Get iterator of OrderedDict."""
        return iter(self.view)

    def __repr__(self):
        """Meaningful repr."""
        return 'View' + repr(list(self.items()))


class Formatter():
    """Format result from pynetbox based on given peramiters."""

    def __init__(self,
                 result,
                 view='table',
                 view_model=None,
                 cols=list(),
                 disable_header=False):
        """Initialize Display instance."""
        self.result = result
        self.view = view
        self.view_model = view_model
        self.cols = cols
        self.disable_header = disable_header
        self._string = ''

    def _build_table(self):

        display = list()

        for i, entry in enumerate(self.result):
            view = self.view_model(entry, cols=self.cols)
            if i == 0:
                display.append([i for i in view.keys()])
            display.append([i for i in view.values()])

        return display

    def _get_detail(self):

        display = list()

        for entry in self.result:
            display.append(self.view_model(entry).detail_view())

        self._string = ('\n\n' + ('#' * 80) + '\n\n').join(display)

    def _get_json(self):

        def build(data):
            if isinstance(data, Record):
                data = dict(data)
            elif isinstance(data, list):
                data = [build(d) for d in data]
            return data

        self._string = json.dumps(build(self.result))

    def _get_table(self):

        display = self._build_table()
        assert len(display) > 1
        if self.disable_header:
            display.pop(0)

        self._string = rend_table(display)

    def _get_view(self):

        if not self.view_model:
            self.view_model = view_name(self.result[0])

        if isinstance(self.view_model, str):

            if self.view_model == 'BaseView':
                self.view_model = BaseView
                return

            view_names = [view.__name__ for view in BaseView.__subclasses__()]

            if self.view_model not in view_names:
                self.view_model = BaseView
                return

            for view in reversed(BaseView.__subclasses__()):
                if view.__name__ == self.view_model:
                    self.view_model = view
                    return

        if isinstance(self.view_model.__class__, BaseView.__class__):
            return

        self.view_model = BaseView

    @property
    def string(self):
        """Generate string based on peramiters passed to Display."""
        if self.view == 'json':
            self._get_json()
            return self._string

        if isinstance(self.result, Record):
            self.result = [self.result]

        if isinstance(self.result, list) and is_list_of_records(self.result):
            self._get_view()
            if self.view == 'detail':
                self._get_detail()
            else:
                self._get_table()
        else:
            self._string = str(self.result)

        return self._string


def nbprint(result,
            view='table',
            view_model=None,
            cols=list(),
            disable_header=False,
            string=False):
    """Print result from pynetbox."""
    formatted_result = Formatter(result,
                                 view=view,
                                 view_model=view_model,
                                 cols=cols,
                                 disable_header=disable_header)
    if string:
        return formatted_result.string
    else:
        print(formatted_result.string)



"""Load sub-commands for nbcli."""


from nbcli.commands.base import BaseSubCommand      # noqa: F401
from nbcli.commands.init import InitSubCommand      # noqa: F401
from nbcli.commands.models import ModelsSubCommand  # noqa: F401
from nbcli.commands.search import SearchSubCommand  # noqa: F401
from nbcli.commands.filter import FilterSubCommand  # noqa: F401
from nbcli.commands.create import CreateSubCommand  # noqa: F401
from nbcli.commands.shell import ShellSubCommand    # noqa: F401



from nbcli.views.tools import BaseView

class DcimSitesView(BaseView):

    def table_view(self):

        self.add_col('Name', self.get_attr('name'))
        self.add_col('Status', self.get_attr('status'))
        self.add_col('Facility', self.get_attr('facility'))
        self.add_col('Region', self.get_attr('region'))
        self.add_col('Tenant', self.get_attr('tenant'))
        self.add_col('ASN', self.get_attr('asn'))
        self.add_col('Description', self.get_attr('description'))



from nbcli.core.utils import app_model_by_loc


class NbArgs():

    def __init__(self, netbox, kwargs=None, action='get'):

        assert action in ['get', 'post', 'patch']
        self._nb = netbox
        self._logger = self._nb.nbcli.logger
        self.args = list()
        self.kwargs = kwargs or {}
        self.action = action

    def __bool__(self):
        return bool(self.args) or bool(self.kwargs)

    def __repr__(self):
        return 'NbArgs(args={}, kwargs={})'.format(self.args, self.kwargs)

    def proc(self, *args):

        for arg in args:
            if isinstance(arg, tuple) and (len(arg) == 2):
                self.update(*arg)
            elif isinstance(arg, str):
                self.string(arg)

    def update(self, key, value):
        if key in self.kwargs:
            if isinstance(self.kwargs[key], list):
                self.kwargs[key].append(value)
            else:
                self.kwargs[key] = [self.kwargs[key], value]
        else:
            self.kwargs[key] = value

    def string(self, string):

        if ':' in string:
            args = list()
            for resol in reversed(string.split('::')):
                al = list()
                for res in resol.split('~'):
                    nba = NbArgs(self._nb, action=self.action)
                    nba.proc(*args)
                    nba.resolve(*res.split(':'), kwargs=nba.kwargs)
                    al += list(nba.kwargs.items())
                args = al
            for arg in args:
                self.update(*arg)
        elif '=' in string:
            kv = string.split('=')
            if (len(kv) == 2) and (len(kv[0]) > 0) and (len(kv[1]) > 0):
                self.update(kv[0], kv[1]) 
            else:
                self._logger.warning("Could not process '%s'", string)
        else:
            self.args.append(string)

    def apply_res(self, result, res):

        rep_items = getattr(res.reply, self.action)

        replyl = list()

        if not rep_items:
            return replyl

        for rep in rep_items:
            replyl += [(rep[0], getattr(obj, rep[1])) for obj in result]

        self.proc(*tuple(replyl))

    def resolve(self, resstr, *args, kwargs=None, res=None):

        res = res or self._nb.nbcli.rm.get(resstr)
        if not res:
            self._logger.warning("Could not resolve '%s'", resstr)
            return

        ep = app_model_by_loc(self._nb, res.model)

        nba = NbArgs(self._nb, kwargs=kwargs)
        nba.proc(*args)
        for arg in nba.args:
            nba.update(res.lookup, arg)

        result = ep.filter(**nba.kwargs)

        self.apply_res(result, res)

        return nba, result



from nbcli.views.tools import BaseView, Formatter

class DcimDevicesView(BaseView):

    def table_view(self):

        self.add_col('Name', self.get_attr('name'))
        self.add_col('Status', self.get_attr('status'))
        self.add_col('Tenant', self.get_attr('tenant'))
        self.add_col('Site', self.get_attr('site'))
        self.add_col('Rack', self.get_attr('rack'))
        self.add_col('Role', self.get_attr('device_role'))
        self.add_col('Type', self.get_attr('device_type'))
        self.add_col('IP Address', str(self.get_attr('primary_ip')).split('/')[0])

    def detail_view(self):

        lines = list()
        lines.append('{}\n{}'.format(self.obj, '=' * len(str(self.obj))))

        lines.append('\nDevice\n------')
        lines.append('{}:\t{}'.format('Site', self.get_attr('site')))
        lines.append('{}:\t{}'.format('Rack', self.get_attr('rack')))
        lines.append('{}:\t{}'.format('Position', self.get_attr('position')))
        lines.append('{}:\t{}'.format('Tenant', self.get_attr('tenant')))
        lines.append('{}:\t{}'.format('Device Type', self.get_attr('device_type')))
        lines.append('{}:\t{}'.format('Serial Number', self.get_attr('serial')))
        lines.append('{}:\t{}'.format('Asset Tag', self.get_attr('asset_tag')))

        lines.append('\nManagement\n----------')
        lines.append('{}:\t{}'.format('Role', self.get_attr('role')))
        lines.append('{}:\t{}'.format('Platform', self.get_attr('platform')))
        lines.append('{}:\t{}'.format('Status', self.get_attr('status')))
        lines.append('{}:\t{}'.format('Primary IPv4', self.get_attr('primary_ip4')))
        lines.append('{}:\t{}'.format('Primary IPv6', self.get_attr('primary_ip6')))
        
        lines.append('\nCustom Fields\n-------------')
        for key, value in self.obj.custom_fields.items():
            lines.append('{}:\t{}'.format(key, value))

        lines.append('\nTags\n----')
        lines.append(' '.join(self.obj.tags))

        lines.append('\nComments\n--------')
        lines.append(self.get_attr('comments'))

        ifaces = self.obj.api.dcim.interfaces.filter(device_id=self.obj.id)
        cols = [('Name', 'name'),
                 ('LAG', 'lag'),
                 ('Description', 'description'),
                 ('MTU', 'mtu'),
                 ('Mode', 'mode'),
                 ('Cable', 'cable.id'),
                 ('Connection', 'connected_endpoint.device'),
                 ('', 'connected_endpoint')]
        if len(ifaces) > 0:
            lines.append('\nInterfaces\n----------')
            lines.append(Formatter(ifaces, cols=cols).string)
        

        return '\n'.join(lines)



from pynetbox.core.response import Record
from nbcli.commands.base import BaseSubCommand
from nbcli.commands.tools import NbArgs
from nbcli.core.utils import app_model_by_loc, is_list_of_records


class Filter():

    def __init__(self,
                 netbox,
                 model,
                 args=list(),
                 count=False,
                 delete=False,
                 ud=list(),
                 de=list()):

        self.model = app_model_by_loc(netbox, model)

        method = 'all'

        if count:
            method = 'count'
        elif args:
            method = 'filter'

        self.method = getattr(self.model, method)

        nba = NbArgs(netbox)
        nba.proc(*args)
        result = self.method(*nba.args, **nba.kwargs)

        if isinstance(result, Record):
            result = [result]

        if is_list_of_records(result):
            if delete:
                self.delete(result)
            elif ud:
                ud_nba = NbArgs(netbox, action='patch')
                ud_nba.proc(*ud)
                self.update(result, ud_nba)
            elif de:
                detail = de.pop(0)
                de_nba = NbArgs(netbox)
                de_nba.proc(*de)
                self.detail(result, detail, *de_nba.args, **de_nba.kwargs)
            else:
                self.result = result
        else:
            self.result = result

    def delete(self, result):
        assert is_list_of_records(result)
        anslist = list()
        for obj in result:
            anslist.append('{} ({})'.format(str(obj), str(obj.id)))
        ans = input('Delete {}?\n* {}\n(yes) to delete: '. \
                    format(obj.__class__.__name__,
                           '\n* '.join(anslist)))
        if ans.lower() == 'yes':
            dellist = list()
            for obj in result:
                objrep = '{} ({})'.format(str(obj), str(obj.id))
                if obj.delete():
                    dellist.append('{} Deleted!'.format(objrep))
                else:
                    dellist.append('Error deleting {}'.format(objrep))
            self.result = '\n'.join(dellist)
        else:
            self.result = 'Aborting!'

    def update(self, result, nba):
        assert is_list_of_records(result)
        anslist = list()
        for obj in result:
            anslist.append('{} ({})'.format(str(obj), str(obj.id)))
        ans = input('Update {} with {}?\n* {}\n(yes) to update: '. \
                    format(obj.__class__.__name__,
                           str(nba.kwargs),
                           '\n* '.join(anslist)))
        if ans.lower() == 'yes':
            udlist = list()
            for obj in result:
                objrep = '{} ({})'.format(str(obj), str(obj.id))
                if obj.update(nba.kwargs):
                    udlist.append('{} Updated!'.format(objrep))
                else:
                    udlist.append('Error updating {}'.format(objrep))
            self.result = '\n'.join(udlist)
        else:
            self.result = 'Aborting!'

    def detail(self, result, detail, *args, **kwargs):
        assert is_list_of_records(result)
        self.result = list()
        for obj in result:
            de = getattr(obj, detail)
            self.result += de.list(*args, **kwargs)


class FilterSubCommand(BaseSubCommand):
    """Filter Netbox objects by searchterm and object properties.
    
    Optionally update and delete objects returned by the filter.
    Control output view and listed columns.
    """

    name = 'filter'
    parser_kwargs = dict(help='Filter NetBox objects.')
    view_options = True


    def setup(self):
    
        self.parser.add_argument('model',
                            help="NetBox model")
    
        self.parser.add_argument('args',
                            nargs='*',
                            help='Argumnet(s) to filter results.')

        obj_meth = self.parser.add_mutually_exclusive_group()
    
        obj_meth.add_argument('-c', '--count',
                              action='store_true',
                              help='Return the count of objects in filter.')

        obj_meth.add_argument('-D', '--delete',
                              action='store_true',
                              help='Delete Object(s) returned by filter [WIP]')
    
        obj_meth.add_argument('--ud', '--update',
                              nargs='*',
                              help='Update object(s) returned by filter '+ \
                                   'with given kwargs [WIP]')
    
        self.parser.add_argument('--de', '--detail-endpoint',
                            nargs='*',
                            help='List results from detail endpoint '+ \
                                 'With optional kwargs [WIP]')


    def run(self):
        """Filter Netbox objects by a searchterm and object properties.

        Usage Examples:

        - Filter IP Addresses with searchterm '192.168.1.1':
          $ nbcli filter address 192.168.1.1

        - Filter devices by serial number using keyword arguments:
          $ nbcli filter device serial=123456

        - Filter devices types by manufacturer using auto-resolve arguments: 
          $ nbcli filter device_type manufacturer:ACME

        - Update tenant on devices returned by filter:
          $ nbcli filter device name=server1 -ud tenant:tenant2

        - Delete IP addresses returned by filter:
          $ nbcli filter address 192.168.1.1 -D
        """
    
        nbfilter = Filter(self.netbox,
                          self.args.model,
                          args=self.args.args or [],
                          count=self.args.count,
                          delete=self.args.delete,
                          ud=self.args.ud or [],
                          de=self.args.de or [])

        if nbfilter.result:
            self.nbprint(nbfilter.result, cols=self.args.cols)
        else:
            self.logger.warning('No results found')



"""Define Classes and Functions used throughout nbcli."""

from collections import namedtuple
import json
import logging
import os
from pathlib import Path
from pynetbox.core.endpoint import Endpoint
from pynetbox.core.response import Record


Resolve = namedtuple('Resolve', ['model', 'alias', 'lookup', 'reply'])
Reply = namedtuple('Reply', ['get', 'post', 'patch'])


class ResMgr():
    """Store and retrieve Resolve objects."""

    def __init__(self, *args, **kwargs):
        """Populate ResMgr with Resolve objects based on args/kwargs."""
        self._res = None
        self._resl = []

        if args:
            self._res = Resolve(*args)

        for key, value in kwargs.items():

            value = value or {}

            if isinstance(value, dict):
                self._proc_res_data(key, value)
            elif isinstance(value, list):
                for data in value:
                    self._proc_res_data(key, data)

        self._resl = tuple(self._resl)

    def _proc_res_data(self, key, data):

        model = key
        alias = data.pop('alias', model.strip('s').split('.')[-1])
        lookup = data.pop('lookup', 'name')
        reply = data.pop('reply', {})
        if isinstance(reply, list):
            reply = Reply(tuple([tuple(i) for i in reply]),
                          tuple([tuple(i) for i in reply]),
                          tuple([tuple(i) for i in reply]))
        elif isinstance(reply, dict):
            # tuple-fy
            get = reply.pop('get', (('{}_id'.format(alias), 'id'),))
            post = reply.pop('post', ((alias, 'id'),))
            patch = reply.pop('patch', post)
            reply = Reply(get, post, patch)

        self._resl.append(ResMgr(model, alias, lookup, reply, **data))

    def __repr__(self):
        """."""
        if self._res:
            return str(self._res)
        else:
            rl = list()
            for res in self._resl:
                if res._resl:
                    rl.append('{}*'.format(res.model))
                else:
                    rl.append(res.model)
            return '{}({})'.format('ResMgr', ', '.join(rl))

    def __iter__(self):
        """Iterate through list of child REsolve objects."""
        return self._resl.__iter__()

    def __getattr__(self, key):
        """Return attribute of Resolve object if defined."""
        if (key in Resolve._fields) and self._res:
            return getattr(self._res, key)

        object.__getattribute__(self, key)

    def get(self, string):
        """Return Resolve object if alias or model matches string."""
        for res in self:
            if res.alias == string:
                return res
            if res.model == string:
                return res
        return None


def view_name(obj):
    """Generate view name based on class, url, or endpoint url."""
    assert isinstance(obj, (Record, Endpoint))

    class_name = obj.__class__.__name__
    model_loc = app_model_loc(obj)

    if class_name not in ['Record', 'Endpoint']:
        return model_loc.split('.')[0].title() + class_name + 'View'

    return model_loc.title().replace('_', '').replace('.', '') + 'View'


def rend_table(table):
    """Convert 2D array into printable string."""
    # get max width for each column
    colw = list()
    for col in range(len(table[0])):
        colw.append(max([len(row[col]) for row in table]))

    # build template based on max with for each column
    template = ''
    buff = 2
    for w in colw:
        template += '{:<' + str(w + buff) + 's}'

    return '\n'.join([template.format(*row) for row in table])


def get_nbcli_dir():
    """Return path of nbcli directory."""
    default = Path.home().joinpath('.nbcli')
    return Path(os.environ.get('NBCLI_DIR', str(default)))


def get_nbcli_logger():
    """Return logger object and set level if defined by env var."""
    logging.basicConfig(format="[%(levelname)s](%(name)s): %(message)s")
    logger = logging.getLogger('nbcli')
    env_level = os.environ.get('NBCLI_LOGLEVEL', 'WARNING').upper()
    if hasattr(logging, env_level):
        logger.setLevel(getattr(logging, env_level))
    logger.debug('Log level DEBUG set by enviornment.')
    return logger


def auto_cast(string):
    """Convert True, False, and None strings to their type."""
    assert isinstance(string, str)
    if string.lower() == 'none':
        return None
    if string.lower() == 'true':
        return True
    if string.lower() == 'false':
        return False
    if ('{' in string) or ('[' in string):
        try:
            return json.loads(string)
        except Exception:
            return string
    return string


def app_model_loc(obj):
    """Derive pynetbox App and Endpoint names from url/endpoint.url."""
    assert isinstance(obj, (Record, Endpoint))
    if obj.url:
        url = obj.url
    else:
        url = obj.endpoint.url
    parts = url.replace(obj.api.base_url, '').strip('/').split('/')
    return '.'.join(parts[:2]).replace('-', '_')


def app_model_by_loc(api, loc):
    """Return Endpoint defined by location."""
    assert isinstance(loc, str)
    loc = loc.lower().replace('-', '_')
    res = api.nbcli.rm.get(loc)
    if res:
        loc = res.model
    app_ep = loc.split('.')
    assert len(app_ep) == 2
    app = getattr(api, app_ep[0])
    ep = getattr(app, app_ep[1])
    return ep


def is_list_of_records(result):
    """Determine if list only contains pynetbox Records of the same type."""
    if isinstance(result, list) and (len(result) > 0):

        # check all entries are an instance of Record
        rc = [isinstance(e, Record) for e in result].count(True) == len(result)

        # check all entries are the same type
        tc = len(set(type(e) for e in result)) == 1

        return rc and tc

    else:
        return False


def getter(obj, string):
    """Get attribute or item from object defined by string."""
    assert isinstance(string, str)
    def getitem(o, k):

        try:
            try:
                return o[str(k)]
            except Exception:
                return o[int(k)]
        except Exception:
            return None

    for attr in string.split('.'):
        keys = attr.split(':')[1:]
        attr = attr.split(':')[0]
        try:
            obj = getattr(obj, attr)
            for key in keys:
                obj = getitem(obj, key)
        except Exception:
            return None
    return obj



"""Objects related to loading nbcli configuration."""

import os
from pkg_resources import resource_string
import pynetbox
import requests
import urllib3
import yaml
from nbcli import logger
from nbcli.core.utils import ResMgr, auto_cast, get_nbcli_dir


class Config():
    """nbcli config Namespace."""

    def __init__(self, init=False):
        """Create Config object.

        Args:
            init (bool): Seting True will create a new configuration file.
        """
        uf = type('tree', (), {})()

        uf.dir = get_nbcli_dir()
        uf.user_config = uf.dir.joinpath('user_config.yml')
        uf.extdir = uf.dir.joinpath('user_extensions')
        uf.user_commands = uf.extdir.joinpath('user_commands.py')
        uf.user_views = uf.extdir.joinpath('user_views.py')

        self.user_files = uf

        if init:
            self._init()
            return

        self._load()

    def _init(self):
        """Create a new empty config file."""
        # Create user directory tree
        dirlist = [self.user_files.dir, self.user_files.extdir]
        for udir in dirlist:
            if udir.exists() and not udir.is_dir():
                logger.critical('%s exists, but is not a directory',
                                str(udir.absolute()))
                raise FileExistsError(str(udir.absolute()))
            else:
                udir.mkdir(exist_ok=True)

        # Create user files
        filelist = [self.user_files.user_config,
                    self.user_files.user_commands,
                    self.user_files.user_views]
        for ufile in filelist:
            if ufile.exists():
                logger.info('%s already exists. Skiping.', str(ufile))
            else:
                ufile.touch()
                default = ufile.name + '.default'
                logger.debug(default)
                with open(str(ufile), 'w') as fh:
                    fh.write(resource_string('nbcli.user_defaults',
                                             default).decode())

        print("Edit pynetbox 'url' and 'token' entries in user_config.yml:")
        print('\t{}'.format(str(self.user_files.user_config.absolute())))

    def _load(self):
        """Set attributes from configfile or os environment variables."""
        conffile = self.user_files.user_config
        try:
            user_config = yaml.safe_load(open(str(conffile)))
        except Exception as e:
            logger.critical('Error loading user_config!')
            logger.critical("Run: 'nbcli init' to create a user_config file")
            raise e

        for key, value in user_config.items():
            if isinstance(value, (dict, type(None))):
                setattr(self, key, value or {})

            # get envars
            prefix = 'NBCLI_{}_'.format(key.upper())

            def has_prefix(ek):
                return ek.find(prefix) == 0

            for envkey in filter(has_prefix, os.environ.keys()):
                attr = envkey[len(prefix):].lower()
                envval = auto_cast(os.environ.get(envkey))
                getattr(self, key)[attr] = envval


def get_session(init=False):
    """Create and return pynetbox api object."""
    conf = Config(init=init)
    delattr(conf, 'user_files')

    if init:
        return

    url = conf.pynetbox['url']
    del conf.pynetbox['url']

    nb = pynetbox.api(url, **conf.pynetbox)
    del conf.pynetbox

    if hasattr(conf, 'requests'):
        reqconf = getattr(conf, 'requests')
        session = requests.Session()
        for key, value in reqconf.items():
            setattr(session, key, value)

        nb.http_session = session
        del conf.requests

    if nb.http_session.verify is False:
        urllib3.disable_warnings()

    nb.nbcli = type('nbcli', (), {})()

    resstr = resource_string('nbcli.core', 'resolve_reference.yml').decode()
    resdict = yaml.safe_load(resstr)
    nb.nbcli.rm = ResMgr(**resdict)

    nb.nbcli.logger = logger
    nb.nbcli.conf = conf

    return nb



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = dict()

with open("nbcli/__version__.py", "r") as fh:
    exec(fh.read(), dict(), version)

setuptools.setup(
    name="nbcli",
    version=version["__version__"],
    author="Eric Geldmacher",
    author_email="egeldmacher@wustl.edu",
    description="CLI for netbox using pynetbox module",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ericgeldmacher/nbcli",
    license="GPLv3",
    packages=setuptools.find_packages(),
    package_data={
        'nbcli': ['user_defaults/*.default', 'core/resolve_reference.yml']
    },
    install_requires=["pynetbox>=5.0.3,<6", 'pyyaml'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "nbcli=nbcli.cli:main",
        ],
    },
)



"""Sub Command to set up bare nbcli user directory."""


import logging
from nbcli.commands.base import BaseSubCommand
from nbcli.core.config import get_session


class InitSubCommand(BaseSubCommand):
    """Initialize nbcli.

    Default confg directory location $HOME/.nbcli
    After running edit $HOME/.nbcli/user_config.yml with your credentials.
    """

    name = 'init'
    parser_kwargs = dict(help='Initialize nbcli.')

    def _pre_run_(self, args):
        """Override BaseSubCommand _pre_run_ method for init special case."""
        self.args = args
        self.logger = logging.getLogger('nbcli.'+self.name)
        self._set_log_level_()
        self.logger.debug(args)
        self.run()

    def run(self):
        """Create nbcli config directory and related files.

        Example Usage:

        - Initialize nbcli
          $ nbcli init
        """
        self.netbox = get_session(init=True)



import pkgutil
import sys
import pynetbox
from pynetbox.core.endpoint import Endpoint
from pynetbox.core.response import Record
from pynetbox.core.query import Request
from nbcli.commands.base import BaseSubCommand
from nbcli.core.utils import app_model_loc, app_model_by_loc
from nbcli.views.tools import nbprint

class Shell():

    def __init__(self, netbox,
                 interactive_shell=None,
                 script=None,
                 cmd=None,
                 interact=False,
                 logger=None):

        if pkgutil.find_loader('IPython') is None:
            interactive_shell = 'python'

        self.interactive_shell = interactive_shell
        self.script = script
        self.cmd = cmd
        self.interact = interact
        self.netbox = netbox
        self.logger = logger
        self.banner = ''
        versions = 'Python {}.{}.{} | NetBox {} | pynetbox {}'
        self.banner += versions.format(*sys.version_info[:3],
                                       self.netbox.version,
                                       pynetbox.__version__)
        self.banner += '\nRoot pynetbox API object: Netbox'
        self.banner += '\nAdditional utilities available:\n\t'
        self.banner += 'lsmodels(), nbprint(), nblogger'
        self.build_ns() 


    def build_ns(self):

        self.ns = dict(Netbox=self.netbox,
                       nblogger=self.logger,
                       nbprint=nbprint)

        for res in self.netbox.nbcli.rm:
            name = res.alias.title().replace('_', '')
            self.ns[name] = app_model_by_loc(self.netbox, res.model)


        def lsmodels():

            for k, v in self.ns.items():
                if isinstance(v, Endpoint):
                    print(k)

        self.ns['lsmodels'] = lsmodels


    def python(self):
        from code import interact, InteractiveConsole

        try:
            import readline
            import rlcompleter
            readline.set_completer(rlcompleter.Completer(self.ns).complete)
            readline.parse_and_bind('tab:complete')
        except:
            pass

        console = InteractiveConsole(locals=self.ns)
        if self.cmd:
            console.runcode(self.cmd)
        elif self.script:
            console.runcode(open(self.script).read())
            if self.interact:
                console.interact(banner='')
        else:
            console.interact(banner=self.banner)
    
    def ipython(self):
        from IPython import start_ipython, __version__
        from traitlets.config.loader import Config
    
        banner = 'IPython {} | '.format(__version__) + self.banner

        c = Config()
        c.TerminalInteractiveShell.banner1 = banner
        argv=[]
        if self.cmd:
            argv.append('-c')
            argv.append(self.cmd)
        elif self.script:
            if self.interact:
                c.TerminalInteractiveShell.banner1 = ''
                argv.append('-i')
            argv.append(self.script)

        start_ipython(argv=argv, user_ns=self.ns, config=c)
    
    def run(self):
        if self.interactive_shell == 'ipython':
            self.ipython()
        else:
            self.python()


class ShellSubCommand(BaseSubCommand):
    """Launch Interactive Shell with pynetbox objects preloaded."""

    name = 'shell'
    parser_kwargs = dict(help='Launch interactive shell')

    def setup(self):

        self.parser.add_argument('script', nargs='?', type=str, help='Script to run')
        self.parser.add_argument('-c', metavar='cmd', type=str, help='Program passed in as string')
        self.parser.add_argument('-i', action='store_true',
                      help='inspect interactively after running script')
        self.parser.add_argument('-s', '--interactive-shell', choices=['python', 'ipython'],
                      default='ipython',
                      help='Specifies interactive shell to use')

    def run(self):
        """Run Shell enviornment.

        Example usage:
        $ nbcli shell -i myscript.py
        $ nbcli shell -s python"""
 
        shell = Shell(self.netbox,
                      interactive_shell=self.args.interactive_shell,
                      script=self.args.script,
                      cmd=self.args.c,
                      interact=self.args.i,
                      logger=self.logger)
        shell.run()



"""Main nbcli app."""

from argparse import ArgumentParser, RawTextHelpFormatter
import sys
from textwrap import dedent
from nbcli import logger
from nbcli.core.extend import load_extensions
from nbcli.commands.base import BaseSubCommand


class CLI():
    """Command Line Interface for Netbox."""

    def __init__(self):
        """Set up argument parser and load sub-commands."""
        epilog = '''
               General Options:
                 -h, --help           show this help message and exit
                 -v, --verbose        Show more logging messages
                 -q, --quiet          Show fewer logging messages'''

        self.parser = ArgumentParser(prog='nbcli',
                                     description='Extensible CLI for Netbox',
                                     epilog=dedent(epilog),
                                     formatter_class=RawTextHelpFormatter)
        self.parser.set_defaults(func=None)

        if 'init' not in sys.argv:
            load_extensions()

        subparsers = self.parser.add_subparsers(title='Commands',
                                                metavar='<command>')

        for command in BaseSubCommand.__subclasses__():
            logger.debug('Loading %s', command.__module__)
            command(subparsers)

    def run(self, argv):
        """Run cli app with the given arguments."""
        args = self.parser.parse_args(argv)

        if args.func:
            args.func(args)
        else:
            self.parser.print_help()


def main():
    """Run main cli app with sys.argv from command line."""
    app = CLI()
    app.run(sys.argv[1:])


if __name__ == '__main__':
    main()
